"""Classes to represent reinforcement bars, layers of reinforcement bars
and groups of reinforcement layers"""

from enum import IntEnum
from math import pi, sin, cos, isclose, copysign

from typing import Tuple, List, Union, Dict, Optional, Any

from abc import ABC, abstractmethod
import numpy as np

from rcdesign.is456 import ecu
from rcdesign.is456.concrete import Concrete
from rcdesign.is456.stressblock import LSMStressBlock

# Rebar Enumerations


class RebarType(IntEnum):
    REBAR_UNDEFINED = 0
    REBAR_MS = 1
    REBAR_HYSD = 2
    REBAR_CUSTOM = 3


RebarLabel = {
    RebarType.REBAR_UNDEFINED: "Undefined",
    RebarType.REBAR_MS: "Mild Steel",
    RebarType.REBAR_HYSD: "HYSD",
    RebarType.REBAR_CUSTOM: "Custom type",
}


class StressType(IntEnum):
    STRESS_NEUTRAL = 0
    STRESS_COMPRESSION = 1
    STRESS_TENSION = 2


StressLabel = {
    StressType.STRESS_NEUTRAL: "Neutral",
    StressType.STRESS_COMPRESSION: "Compression",
    StressType.STRESS_TENSION: "Tension",
}


class ShearRebarType(IntEnum):
    SHEAR_REBAR_VERTICAL_STIRRUP = 1
    SHEAR_REBAR_INCLINED_STIRRUP = 2
    SHEAR_REBAR_BENTUP_SINGLE = 3
    SHEAR_REBAR_BENTUP_SERIES = 4


ShearRebarLabel = {
    ShearRebarType.SHEAR_REBAR_VERTICAL_STIRRUP: "Vertical stirrup",
    ShearRebarType.SHEAR_REBAR_INCLINED_STIRRUP: "Inclined stirrup",
    ShearRebarType.SHEAR_REBAR_BENTUP_SINGLE: "Bent-up bars in single group",
    ShearRebarType.SHEAR_REBAR_BENTUP_SERIES: "Bent-up bars in series",
}

# Rebar class


class Rebar(ABC):  # pragma: no cover
    def __init__(
        self,
        label: str,
        fy: float,
        gamma_m: float = 1.15,
        density: float = 78.5,
        Es: float = 2e5,
    ):
        self.label = label
        self.fy = fy
        self.gamma_m = gamma_m
        self.density = density
        self.Es = Es
        self.rebar_type = RebarType.REBAR_UNDEFINED

    @property
    def fd(self) -> float:
        return self.fy / self.gamma_m

    def es_min(self) -> float:
        return self.fd / self.Es + 0.002

    @abstractmethod
    def fs(self, es: float) -> float:
        pass


"""Mild steel reinforcement bars as defined in IS456:2000 with a
well defined yield point"""


class RebarMS(Rebar):
    def __init__(self, label: str, fy: float):
        super().__init__(label, fy)
        self.rebar_type = RebarType.REBAR_MS

    def __repr__(self):
        return f"{self.label:>6}: Type={RebarLabel[self.rebar_type]} fy={self.fy} fd={self.fd:.2f}"

    def fs(self, es: float) -> float:
        _es = abs(es)
        _esy = self.fd / self.Es

        if _es < _esy:
            return es * self.Es
        else:
            return copysign(self.fd, es)


"""High yield strength deformed bars as defined in IS456:2000 with piece-wise
linear stress-strain relation between 0.8 to 1.0 times design strength"""


class RebarHYSD(Rebar):
    inel = np.array(
        [
            [0.8, 0.85, 0.9, 0.95, 0.975, 1.0],
            [0.0, 0.0001, 0.0003, 0.0007, 0.001, 0.002],
        ]
    ).T

    def __init__(self, label: str, fy: float):
        super().__init__(label, fy)
        self.rebar_type = RebarType.REBAR_HYSD
        self.es = RebarHYSD.inel.copy()
        self.es[:, 0] = self.es[:, 0] * self.fy / self.gamma_m
        self.es[:, 1] = self.es[:, 0] / self.Es + self.es[:, 1]

    def __repr__(self) -> str:
        return f"{self.label:>6}: Type={RebarLabel[self.rebar_type]} fy={self.fy} fd={self.fd:.2f}"

    def fs(self, es: float) -> float:
        _es = abs(es)
        i = 0
        fd1 = self.es[i, 0]
        es1 = self.es[i, 1]
        if _es <= es1:
            return es * self.Es
        else:
            for i in range(1, 6):
                fd2 = self.es[i, 0]
                es2 = self.es[i, 1]
                if _es <= es2:
                    break
                else:
                    fd1, es1 = fd2, es2
            if _es < es2:
                fs = fd1 + (fd2 - fd1) / (es2 - es1) * (_es - es1)
            else:
                fs = fd2
            return copysign(fs, es)


"""Layer of reinforcement bars"""


class RebarLayer:
    def __init__(self, rebar: Rebar, dia: List[int], _dc: float):
        self.rebar = rebar
        self.dia = dia
        self._dc = _dc
        self._xc: float = _dc
        self._stress_type: StressType = StressType.STRESS_NEUTRAL

    @property
    def max_dia(self) -> int:
        return max(self.dia)

    @property
    def area(self) -> float:
        return sum([d**2 for d in self.dia]) * pi / 4

    @property
    def dc(self) -> float:
        return self._dc

    @dc.setter
    def dc(self, _dc) -> None:
        self._dc = _dc
        if _dc >= 0:
            self._xc = _dc

    @property
    def xc(self) -> float:
        return self._xc

    @xc.setter
    def xc(self, D: float) -> float:
        if self.dc > 0:
            self._xc = self.dc
        else:
            self._xc = D + self.dc
        return self._xc

    def stress_type(self, xu: float) -> StressType:
        if isclose(self._xc, xu):
            self._stress_type = StressType.STRESS_NEUTRAL
        elif self._xc < xu:
            self._stress_type = StressType.STRESS_COMPRESSION
        elif self._xc > 0:
            self._stress_type = StressType.STRESS_TENSION
        return self._stress_type

    def x(self, xu: float) -> float:
        return xu - self._xc

    def __repr__(self) -> str:
        s = "Dia: "
        b = ""
        for bardia in self.dia:
            b += f"{bardia}, "
        b = "[" + b[:-2] + "]"
        s += f"{b} at {self.dc}. Area: {self.area:.2f} (xc = {self._xc:.2f})"
        return s

    def es(self, xu: float, ecmax: float = ecu) -> float:
        return ecmax / xu * self.x(xu)

    def fs(self, xu: float, ecmax: float = ecu) -> float:
        return self.rebar.fs(self.es(xu, ecmax))

    def force_compression(
        self,
        xu: float,
        csb: LSMStressBlock,
        conc: Concrete,
        ecmax: float = ecu,
    ) -> Tuple[float, float, Dict]:
        x = self.x(xu)
        esc = self.es(xu, ecmax)
        fsc = self.rebar.fs(esc)  # Stress in compression steel
        fcc = conc.fd * csb._fc_(esc)  # Stress in concrete
        _f = self.area * (fsc - fcc)
        _m = _f * x
        result = {"x": x, "esc": esc, "f_s": fsc, "f_c": fcc, "C": _f, "M": _m}
        return _f, _m, result

    def force_tension(self, xu: float, ecmax: float = ecu) -> Tuple[float, float, Dict]:
        x = abs(self.x(xu))
        est = ecmax / xu * x
        fst = self.rebar.fs(est)

        _f = self.area * fst
        _m = _f * x
        result = {"x": x, "est": est, "f_st": fst, "T": _f, "M": _m}
        return _f, _m, result

    def report(
        self,
        xu: float,
        csb: LSMStressBlock,
        conc: Concrete,
        rebar: Rebar,
        ecmax: float = ecu,
    ) -> Dict[str, Any]:  # pragma: no cover
        x = self.x(xu)

        if x >= 0:  # Compression
            esc = ecmax / xu * x
            fsc = rebar.fs(esc)
            fcc = csb.fc(x / xu, ecmax) * conc.fd
            _f = self.area * (fsc - fcc)
            _m = _f * x
            d = {
                "type": "C",
                "dc": self._dc,
                "xc": self._xc,
                "bars": self.bar_list(),
                "area": self.area,
                "x": x,
                "es": esc,
                "fs": fsc,
                "fc": fcc,
                "F": _f,
                "M": _m,
            }
        else:  # Tension
            est = ecmax / xu * abs(x)
            fst = rebar.fs(est)

            _f = self.area * fst
            _m = _f * abs(x)
            d = {
                "type": "T",
                "dc": self._dc,
                "xc": self._xc,
                "bars": self.bar_list(),
                "area": self.area,
                "x": x,
                "es": est,
                "fs": fst,
                "fc": 0,
                "F": _f,
                "M": _m,
            }
        return d

    def bar_list(self, sep=";") -> str:
        from collections import Counter

        d = Counter(self.dia)
        s = ""
        for k in sorted(d):
            s += f"{d[k]}-{k} "
        s = s.rstrip().replace(" ", sep)
        return s

    def __lt__(self, b) -> bool:
        return self._xc < b._xc

    def __le__(self, b) -> bool:
        return self._xc <= b._xc

    def __eq__(self, b) -> bool:
        return self._xc == b._xc

    def __ne__(self, b) -> bool:
        return self._xc != b._xc

    def __gt__(self, b) -> bool:
        return self._xc > b._xc

    def __ge__(self, b) -> bool:
        return self._xc >= b._xc

    def spacing(self, b: float, clear_cover: float) -> float:
        return (b - (2 * clear_cover) - sum(self.dia)) / (len(self.dia) - 1)

    def asdict(
        self, xu: float, sb: LSMStressBlock, conc: Concrete, ecmax: float = ecu
    ) -> dict:
        x = self.x(xu)
        es = ecmax / xu * x
        fsc = self.rebar.fs(es)
        fcc = sb._fc_(es) * conc.fd
        F = self.area * (fsc - fcc)
        d = {
            "fy": self.rebar.fy,
            "Bars": self.bar_list(),
            "xc": self.xc,
            "Strain": es,
            "Type": StressLabel[self.stress_type(xu)][0].capitalize(),
            "f_s": fsc,
            "f_c": fcc,
            "F": F,
            "M": F * x,
        }
        return d


"""Group of reinforcement bars"""


class RebarGroup:
    def __init__(self, layers: List[RebarLayer]):
        self.layers = layers  # List of layers of bars, in any order from edge

    @property
    def area(self) -> float:
        return sum([L.area for L in self.layers])

    def calc_xc(self, D: float) -> None:
        for L in self.layers:
            L.xc = D
        return None

    def centroid(self, xu: float) -> Tuple[float, float]:
        a1 = m1 = a2 = m2 = 0.0

        for L in self.layers:
            if L.stress_type(xu) == StressType.STRESS_COMPRESSION:
                a1 += L.area
                m1 += L.area * L.xc
            elif L.stress_type(xu) == StressType.STRESS_TENSION:
                a2 += L.area
                m2 += L.area * L.xc
        x1 = x2 = 0.0
        if a1 > 0:
            x1 = m1 / a1
        if a2 > 0:
            x2 = m2 / a2
        return x1, x2

    def has_comp_steel(self, xu: float) -> bool:
        for L in self.layers:
            if L._xc < xu:
                return True
        return False

    def area_comp(self, xu: float) -> float:
        a = 0.0
        for L in self.layers:
            if L._xc < xu:
                a += L.area
        return a

    def area_tension(self, xu: float) -> float:
        a = 0.0
        for L in self.layers:
            if L._xc > xu:
                a += L.area
        return a

    def calc_stress_type(self, xu: float) -> None:
        for L in self.layers:
            L.stress_type(xu)

    def force_moment(
        self,
        xu: float,
        csb: LSMStressBlock,
        conc: Concrete,
        ecmax: float = ecu,
    ) -> Tuple[float, float, float, float]:
        fc, mc = self.force_compression(xu, csb, conc, ecmax)
        ft, mt = self.force_tension(xu, ecmax)
        return fc, mc, ft, mt

    def force_tension(self, xu: float, ecmax: float = ecu) -> Tuple[float, float]:
        f = m = 0.0
        for L in self.layers:
            if L._xc > xu:
                # D_xu = L._xc - xu
                _f, _m, _ = L.force_tension(xu, ecmax)
                f += _f
                m += _m
        return f, m

    def force_compression(
        self,
        xu: float,
        csb: LSMStressBlock,
        conc: Concrete,
        ecmax: float = ecu,
    ) -> Tuple[float, float]:
        _f = _m = 0.0
        for L in self.layers:
            if L._xc < xu:
                __f, __m, _ = L.force_compression(xu, csb, conc, ecmax)
                _f += __f
                _m += __m
        return _f, _m

    def __repr__(self) -> str:
        sl = "layers" if len(self.layers) > 1 else "layer"
        s = f"{len(self.layers)} {sl}\n"
        s += f"{'dc':>10}{'xc':>10}{'Bars':>12}{'Area':>10}\n"
        for L in sorted(self.layers):
            s += f"{L._dc:10.2f}{L._xc:10.2f}{L.bar_list():>12}{L.area:10.2f}{StressLabel[L._stress_type].capitalize():>15}\n"
        s += " " * 32 + "-" * 10 + "\n"
        s += f"{self.area:42.2f}\n"
        return s

    def report(
        self,
        xu: float,
        csb: LSMStressBlock,
        conc: Concrete,
        ecmax: float,
    ) -> List[Dict[str, Union[str, int, float]]]:  # pragma: no cover
        result = []
        for L in sorted(self.layers):
            result.append(L.report(xu, csb, conc, L.rebar, ecmax))
        return result


"""Shear reinforcement"""


class ShearReinforcement(ABC):  # pragma: no cover
    def __init__(self, rebar: Rebar, _Asv: float = 0.0, _sv: float = 0.0):
        self.rebar = rebar
        self.__Asv = _Asv
        self._sv = _sv

    @abstractmethod
    def _Asv(self):
        pass

    @abstractmethod
    def Vus(self, d: float) -> float:
        pass

    @abstractmethod
    def get_type(self):
        pass

    @abstractmethod
    def report(self, d: float) -> dict:
        pass


"""Vertical or inclined stirrups as shear reinforcement"""


class Stirrups(ShearReinforcement):
    def __init__(
        self,
        rebar: Rebar,
        _nlegs: int,
        _bar_dia: int,
        _sv: float = 0.0,
        _alpha_deg: float = 90,
    ):
        super().__init__(rebar, _nlegs * pi * _bar_dia**2 / 4, _sv)
        self._nlegs = _nlegs
        self._bar_dia = _bar_dia
        self._alpha_deg = _alpha_deg

    def _Asv(self) -> float:
        self.__Asv = self.nlegs * pi * self.bar_dia**2 / 4
        return self.__Asv

    @property
    def Asv(self):
        return self._Asv()

    @property
    def nlegs(self) -> int:
        return self._nlegs

    @nlegs.setter
    def nlegs(self, n) -> float:
        self._nlegs = n
        self.__Asv = self.nlegs * pi * self.bar_dia**2 / 4
        return self.__Asv

    @property
    def bar_dia(self) -> int:
        return self._bar_dia

    @bar_dia.setter
    def bar_dia(self, dia) -> None:
        self._bar_dia = dia
        self.__Asv = self.nlegs * pi * self.bar_dia**2 / 4

    @property
    def sv(self) -> float:
        return self._sv

    @sv.setter
    def sv(self, _sv: float) -> float:
        self._sv = _sv
        return self._sv

    def calc_sv(self, Vus: float, d: float) -> Optional[float]:
        if (self._alpha_deg < 45) or (self._alpha_deg > 90):
            return None
        alpha_rad = self._alpha_deg * pi / 180
        self._sv = self.rebar.fd * self.Asv * d * sin(alpha_rad) / Vus
        return self._sv

    def __repr__(self) -> str:
        sh_rein = "Vertical" if self._alpha_deg == 90 else "Inclined"
        s = f"{sh_rein} Stirrups: {self.rebar.label} "
        s += f"{self._nlegs}-{self._bar_dia} @ {self._sv} c/c"
        if self._alpha_deg != 90:
            s += " inclined at {self._alpha_deg} degrees"
        s += f" (Asv = {self.Asv:.2f})"
        return s

    def Vus(self, d: float) -> float:
        V_us = self.rebar.fd * self.Asv * d / self._sv
        if self._alpha_deg != 90:
            alpha_rad = self._alpha_deg * pi / 180
            V_us *= sin(alpha_rad) + cos(alpha_rad)
        return V_us

    def get_type(self) -> int:
        if self._alpha_deg == 90:
            return ShearRebarType.SHEAR_REBAR_VERTICAL_STIRRUP
        else:
            return ShearRebarType.SHEAR_REBAR_INCLINED_STIRRUP

    def report(self, d: float) -> dict:
        data = {
            "sh_type": self.get_type(),
            "label": "Stirrups",
            "type": "Vertical",
            "fy": self.rebar.fy,
            "bar_dia": self.bar_dia,
            "legs": self.nlegs,
            "sv": self._sv,
            "alpha": self._alpha_deg,
            "Asv": self.Asv,
            "Vus": self.Vus(d),
        }
        return data


"""Single group (sv == 0) or a series of parallel (sv != 0) bent-up bars as shear reinforcement"""


class BentupBars(ShearReinforcement):
    def __init__(
        self, rebar: Rebar, bars: List[int], _alpha_deg: float = 45, _sv: float = 0.0
    ):
        super().__init__(rebar, pi / 4 * sum([x**2 for x in bars]), _sv)
        self.bars = bars
        self._alpha_deg = _alpha_deg

    def _Asv(self) -> float:
        area = 0.0
        for bar_dia in self.bars:
            area += bar_dia**2
        self.__Asv = pi / 4 * area
        return self.__Asv

    @property
    def Asv(self):
        return self._Asv()

    def Vus(self, d: float = 0.0) -> float:
        V_us = self.rebar.fd * self.Asv
        alpha_rad = self._alpha_deg * pi / 180
        if self._sv == 0:  # Single group of parallel bars
            V_us *= sin(alpha_rad)
        else:  # Series of bars bent-up at different sections
            V_us *= d / self._sv * (sin(alpha_rad) + cos(alpha_rad))
        return V_us

    def get_type(self) -> int:
        if self._sv == 0:
            return ShearRebarType.SHEAR_REBAR_BENTUP_SINGLE
        else:
            return ShearRebarType.SHEAR_REBAR_BENTUP_SERIES

    def __repr__(self) -> str:
        if self._sv == 0:
            s = f"Single group of parallel bent-up bars: {self.rebar.label}-[{self.bars}]"
        else:
            s = f"Series of parallel bent-up bars: {self.rebar.label}-{self.bars} @ {self._sv} c/c"
        if self._alpha_deg != 90:
            s += f" inclined at {self._alpha_deg} degrees"
        s += f" (Asv = {self.Asv:.2f})"
        return s

    def report(self, d: float) -> dict:
        if self._sv == 0:
            bupbar_type = "Single group"
        else:
            bupbar_type = "Series"
        data = {
            "sh_type": self.get_type(),
            "label": "Bentup bars",
            "type": bupbar_type,
            "fy": self.rebar.fy,
            "bars": self.bars,
            "sv": self._sv,
            "alpha": self._alpha_deg,
            "Asv": self.Asv,
            "Vus": self.Vus(d),
        }
        return data


class ShearRebarGroup:
    def __init__(self, shear_reinforcement: List[ShearReinforcement]):
        self.shear_reinforcement = shear_reinforcement

    def Asv(self) -> List[float]:
        asv = []
        for reinf in self.shear_reinforcement:
            asv.append(reinf._Asv())
        return asv

    def Vus(self, d: float) -> List[float]:
        vus = []
        for reinf in self.shear_reinforcement:
            vus.append(reinf.Vus(d))
        return vus

    def get_type(self) -> Dict[ShearRebarType, int]:
        d = {
            ShearRebarType.SHEAR_REBAR_VERTICAL_STIRRUP: 0,
            ShearRebarType.SHEAR_REBAR_INCLINED_STIRRUP: 0,
            ShearRebarType.SHEAR_REBAR_BENTUP_SINGLE: 0,
            ShearRebarType.SHEAR_REBAR_BENTUP_SERIES: 0,
        }
        for reinf in self.shear_reinforcement:
            if reinf.get_type() == ShearRebarType.SHEAR_REBAR_VERTICAL_STIRRUP:
                d[ShearRebarType.SHEAR_REBAR_VERTICAL_STIRRUP] += 1
            elif reinf.get_type() == ShearRebarType.SHEAR_REBAR_INCLINED_STIRRUP:
                d[ShearRebarType.SHEAR_REBAR_INCLINED_STIRRUP] += 1
            elif reinf.get_type() == ShearRebarType.SHEAR_REBAR_BENTUP_SINGLE:
                d[ShearRebarType.SHEAR_REBAR_BENTUP_SINGLE] += 1
            elif reinf.get_type() == ShearRebarType.SHEAR_REBAR_BENTUP_SERIES:
                d[ShearRebarType.SHEAR_REBAR_BENTUP_SERIES] += 1
        return d

    def __repr__(self):
        s = ""
        for sh_reinf in self.shear_reinforcement:
            s += f"{sh_reinf}\n"
        return s


class LateralTie:
    def __init__(self, rebar: Rebar, bar_dia: int, spacing: float):
        self.rebar = rebar
        self.bar_dia = bar_dia
        self.spacing = spacing

    def __repr__(self):
        s = f"Lateral Ties: {self.rebar} - {self.bar_dia}"
        if self.n > 1:
            s += " {self.n}"
        return s



from rcdesign.is456.stressblock import LSMStressBlock
from rcdesign.is456.concrete import Concrete
from rcdesign.is456.rebar import (
    RebarHYSD,
    RebarLayer,
    RebarGroup,
    ShearRebarGroup,
    Stirrups,
)
from rcdesign.is456.section import RectBeamSection

if __name__ == "__main__":
    sb = LSMStressBlock("LSM Flexure")
    m20 = Concrete("M20", 20)
    fe415 = RebarHYSD("Fe 415", 415)

    t1 = RebarLayer(fe415, [20, 16, 20], -35)
    steel = RebarGroup([t1])
    sh_st = ShearRebarGroup([Stirrups(fe415, 2, 8, 150)])

    sec = RectBeamSection(230, 450, sb, m20, steel, sh_st, 25)
    xu = sec.xu(0.0035)
    print("Example 1\n")
    print(sec.report(xu, 0.0035))
    print(f"{'='*80}\n")

    # Doubly reinforced section
    sb = LSMStressBlock("IS456 LSM")
    m20 = Concrete("M20", 20)
    fe415 = RebarHYSD("Fe 415", 415)

    c1 = RebarLayer(fe415, [16, 16], 35)
    t1 = RebarLayer(fe415, [16, 16, 16], -35)
    t2 = RebarLayer(fe415, [16, 16], -70)
    steel = RebarGroup([c1, t1, t2])
    sh_st = ShearRebarGroup([Stirrups(fe415, 2, 8, 150)])

    sec = RectBeamSection(230, 450, sb, m20, steel, sh_st, 25)
    xu = sec.xu(0.0035)
    print("Example 2\n")
    print(sec.report(xu, 0.0035))



from math import sqrt, pi, ceil
from typing import List


class Beam:
    ecy = 0.002
    ecu = 0.0035
    gamma_mc = 1.5
    gamma_ms = 1.15
    Es = 2e5

    def xumax_d(self, fy: float) -> float:
        return self.ecu / (fy / self.gamma_ms / self.Es + 0.002 + self.ecu)

    def Mulim_const(self, fy: float) -> float:
        xumax_d = self.xumax_d(fy)
        k = (17 / 21) * (0.67 / self.gamma_mc) * xumax_d * (1 - 99 / 238 * xumax_d)
        return k

    def reqd_d(self, fck: float, fy: float, b: float, Mu: float) -> float:
        Mulim_fckbd2 = self.Mulim_const(fy)
        return sqrt(Mu / (Mulim_fckbd2 * fck * b))

    def reqd_xu_d(self, fck: float, b: float, d: float, Mu: float) -> float:
        aa = 1.0
        bb = -238 / 99
        cc = (21 / 17) * (1.5 / 0.67) * (238 / 99) * (Mu / (fck * b * d ** 2))
        xu_d = (-bb - sqrt(bb ** 2 - 4 * aa * cc)) / (2 * aa)
        return xu_d

    def reqd_Ast(self, fck: float, fy: float, b: float, d: float, Mu: float) -> float:
        xu = self.reqd_xu_d(fck, b, d, Mu) * d
        return Mu / ((fy / self.gamma_ms) * (d - (99 / 238 * xu)))

    @staticmethod
    def bar_area(dia: float) -> float:
        return pi / 4 * dia ** 2

    @staticmethod
    def num_bars(ast: float, dia: float) -> int:
        return int(ceil(ast / Beam.bar_area(dia)))

    @staticmethod
    def spacing(b: float, cl_cov: float, bars: List[int]) -> float:
        n = len(bars)
        return (b - (2 * cl_cov) - sum(bars)) / (n - 1)



from math import sqrt
import numpy as np

# Concrete class
"""Concrete class with stress-strain properties as defined in IS456:2000"""


class Concrete:
    def __init__(
        self,
        label: str,
        fck: float,
        gamma_m: float = 1.5,
        density: float = 25.0,
    ):
        self.label = label
        self.fck = fck
        self.gamma_m = gamma_m
        self.density = density

    def __repr__(self) -> str:
        s = f"fck = {self.fck:.2f} N/mm^2, fd = {self.fd:.2f} N/mm^2"
        return s

    @property
    def Ec(self) -> float:
        return 5000 * sqrt(self.fck)

    @property
    def fd(self) -> float:
        return 0.67 * self.fck / self.gamma_m

    def tauc(self, pt: float) -> float:
        if pt < 0.15:
            pt = 0.15
        if pt > 3:
            pt = 3.0
        beta = max(1.0, (0.8 * self.fck) / (6.89 * pt))
        num = 0.85 * sqrt(0.8 * self.fck) * (sqrt(1 + 5 * beta) - 1)
        den = 6 * beta
        return num / den

    def tauc_max(self):
        tauc = np.array([[15, 20, 25, 30, 35, 40], [2.5, 2.8, 3.1, 3.5, 3.7, 4.0]])
        if self.fck < 15:
            return 0.0
        elif self.fck >= tauc[0, -1]:
            return tauc[1, -1]
        else:
            for i in range(1, tauc.shape[1]):
                if self.fck <= tauc[0, i]:
                    if self.fck == tauc[0, i]:
                        return tauc[1, i]
                    else:
                        x1 = tauc[0, i - 1]
                        y1 = tauc[1, i - 1]
                        x2 = tauc[0, i]
                        y2 = tauc[1, i]
                        print(x1, x2, y1, y2)
                        return y1 + (y2 - y1) / (x2 - x1) * (self.fck - x1)



"""Class to represent reinforced concrete cross sections"""

from math import isclose
from enum import Enum
from typing import Tuple, List, Any

# from dataclasses import dataclass
# from abc import ABC, abstractmethod
from scipy.optimize import brentq

from rcdesign.is456 import ecu
from rcdesign.is456.stressblock import LSMStressBlock
from rcdesign.is456.concrete import Concrete
from rcdesign.is456.rebar import (
    RebarGroup,
    ShearRebarGroup,
    LateralTie,
    ShearRebarType,
    StressType,
    StressLabel,
)
from rcdesign.is456.design import Beam
from rcdesign.utils import rootsearch, underline, header


# DesignForce class


class DesignForceType(Enum):
    BEAM = 1
    COLUMN = 2
    SLAB = 3
    SHEARWALL = 4


"""Class to repersent a rectangular beam section"""


class RectBeamSection:
    def __init__(
        self,
        b: float,
        D: float,
        csb: LSMStressBlock,
        conc: Concrete,
        long_steel: RebarGroup,
        shear_steel: ShearRebarGroup,
        clear_cover: float,
    ):
        self.design_force_type = DesignForceType.BEAM
        self.csb = csb
        self.conc = conc
        self.long_steel = long_steel
        self.clear_cover = clear_cover
        self.b = b
        self.D = D

        self.shear_steel = shear_steel
        self.calc_xc()

    def calc_xc(self) -> None:
        self.long_steel.calc_xc(self.D)
        return None

    def calc_stress_type(self, xu: float) -> None:
        self.calc_xc()
        self.long_steel.calc_stress_type(xu)

    def C(self, xu: float, ecmax: float = ecu) -> Tuple[float, float]:
        Fc, Mc, _, _ = self.F_M(xu, ecmax)
        return Fc, Mc

    def T(self, xu: float, ecmax: float) -> Tuple[float, float]:
        _, _, Ft, Mt = self.F_M(xu, ecmax)
        return Ft, Mt

    def C_T(self, xu: float, ecmax: float = ecu) -> float:
        self.calc_stress_type(xu)
        C, _, T, _ = self.F_M(xu, ecmax)
        return C - T

    def F_M(self, xu: float, ecmax: float = ecu) -> Tuple[float, float, float, float]:
        # sb = LSMStressBlock("LSM Flexure")
        self.calc_stress_type(xu)
        Fc = Mc = Ft = Mt = 0.0
        # Compression force - concrete
        k = xu / self.D
        Fcc = self.csb.C(0, k, k, ecmax) * self.conc.fd * self.b * self.D
        Mcc = self.csb.M(0, k, k, ecmax) * self.conc.fd * self.b * self.D**2
        # Compression force - compression steel
        Fsc, Msc, Fst, Mst = self.long_steel.force_moment(
            xu, self.csb, self.conc, ecmax
        )
        # Tension force in tension steel
        Ft, Mt = self.long_steel.force_tension(xu, ecmax)
        Fc = Fcc + Fsc
        Mc = Mcc + Msc
        return Fc, Mc, Ft, Mt

    def xu(self, ecmax: float = ecu) -> float | Any:
        dc_max = 10

        x1, x2 = rootsearch(self.C_T, dc_max, self.D, 10, ecmax)
        x = brentq(self.C_T, x1, x2, args=(ecmax,))
        return x

    def Mu(self, xu: float, ecmax: float = ecu) -> float:
        # Assuming area of tension steel to be such as to produce a tension force equal to C
        _, Mc = self.C(xu, ecmax)
        _, Mt = self.T(xu, ecmax)
        M = Mc + Mt
        return M

    def tauc(self, xu: float) -> float:
        return self.conc.tauc(self.pt(xu))

    def __repr__(self) -> str:
        ecmax = self.csb.ecu
        xu = self.xu(ecmax)
        return self.report(xu, ecmax)

    def has_compr_steel(self, xu: float) -> bool:
        for L in self.long_steel.layers:
            if L._xc < xu:
                return True
        return False

    def report(self, xu: float, ecmax: float = ecu) -> str:  # pragma: no cover
        self.calc_xc()
        self.calc_stress_type(xu)
        k = xu / self.D
        ecy = self.csb.ecy
        hdr0 = f"RECTANGULAR BEAM SECTION: {self.b} x {self.D}"
        s = f"{header(hdr0, '~')}\n"
        s += f"{header('FLEXURE', '=')}\nEquilibrium NA = {xu:.2f} (k = {k:.2f}) (ec_max = {ecmax:.6f})\n\n"
        fcc = self.csb._fc_(ecmax) * self.conc.fd
        Fc = self.b * self.csb.C(0, k, k, ecmax) * self.conc.fd * self.D
        Mc = self.csb.M(0, k, k) * self.conc.fd * self.b * self.D**2
        hdr1 = f"{'fck':>6} {' ':>8} {' ':>12} {'ec_max':>12} {'Type':>4} "
        hdr1 += f"{' ':>8} {'f_c':>6} {'F (kN)':>8} {'M (kNm)':>8}"
        s += hdr1 + "\n" + underline(hdr1) + "\n"
        s += f"{self.conc.fck:6.2f} {' ':>8} {' ':>12} {ecmax:12.8f} {'C':>4} {' ':>8} {fcc:6.2f} "
        s += f"{Fc / 1e3:8.2f} {Mc/ 1e6:8.2f}\n{underline(hdr1)}\n\n"
        Ft = 0.0
        Mt = 0.0
        hdr2 = f"{'fy':>6} {'Bars':>12} {'xc':>8} {'Strain':>12} {'Type':>4} {'f_s':>8} {'f_c':>6}"
        hdr2 += f" {'F (kN)':>8} {'M (kNm)':>8}"
        s += f"{hdr2}\n{underline(hdr2)}\n"
        for L in sorted(self.long_steel.layers):
            z = k - (L._xc / self.D)
            esc = self.csb.ec(z, k) * ecy
            stress_type = L.stress_type(xu)
            fsc = L.rebar.fs(esc)
            s += f"{L.rebar.fy:6.0f} {L.bar_list():>12} {L._xc:8.2f} {esc:12.8f} "
            s += f"{StressLabel[stress_type][0]:>4} {fsc:8.2f} "
            if stress_type == StressType.STRESS_COMPRESSION:
                fcc = self.csb.fc(z, k, ecmax) * self.conc.fd
                c = L.area * (fsc - fcc)
                s += f"{fcc:6.2f} "
            elif L._stress_type == StressType.STRESS_TENSION:
                c = L.area * fsc
                s += f"{' ':>6} "

            m = c * (k * self.D - L._xc)
            s += f"{c/1e3:8.2f} {m/1e6:8.2f}\n"
            Ft += c
            Mt += m
        s += f"{underline(hdr2)}\n"
        if len(self.long_steel.layers) > 1:
            C_M = f"{Ft/1e3:8.2f} {Mt/1e6:8.2f}"
            s += f"{' '*62} {C_M}\n{' '*62} {underline(C_M, '=')}\n"
        F = 0.0 if isclose(Fc + Ft, 0, abs_tol=1e-10) else Fc + Ft
        C_M = f"{F/1e3:8.2f} {(Mc + Mt)/1e6:8.2f}"
        s += f"{' ':>62} {C_M}\n"
        s += f"{header('SHEAR', '=')}\n"
        tauc = self.conc.tauc(self.pt(xu))
        area = self.b * self.eff_d(xu)
        vuc = area * tauc
        hdr3 = f"{'Type':>14} {' ':>14} {'tau_c':>6} {'Area (mm^2)':>16} {' ':>8} {' ':>8} {'V_uc (kN)':>8}"
        s += f"{header(hdr3)}\n"
        s += f"{'Concrete':>14} {' ':>14} {tauc:6.2f} {area:16.2f} {' ':>8} {' ':>8} {vuc/1e3:8.2f}\n"
        s += f"{underline(hdr3)}\n"
        hdr4 = f"{'Type':>14} {'Variant':>14} {'f_y':>6} {'Bars':>16} {'s_v':>8} {'A_sv':>8} {'V_us (kN)':>8}"
        s += f"{header(hdr4)}\n"
        vus = 0.0
        for sh_rein in self.shear_steel.shear_reinforcement:
            data = sh_rein.report(self.eff_d(xu))
            s += f"{data['label']:>14} {data['type']:>14} {data['fy']:6} "
            if data["sh_type"] in [
                ShearRebarType.SHEAR_REBAR_VERTICAL_STIRRUP,
                ShearRebarType.SHEAR_REBAR_INCLINED_STIRRUP,
            ]:
                bar_info = f"{data['legs']}-{data['bar_dia']}#"
            else:
                bar_info = f"{data['bars']}"
            s += f"{bar_info:>16} {data['sv']:8.1f} {data['Asv']:8.2f} {data['Vus']/1e3:8.2f}\n"
            vus += data["Vus"]
        vu = f"{(vuc + vus)/1e3:8.2f}"
        s += f"{' ':>71} {underline(vu, '=')}\n{' ':>71} {vu}\n"
        s += (
            f"{header('CAPACITY', '=')}\n{'Mu = ':>5}{self.Mu(xu, ecmax)/1e6:.2f} kNm\n"
        )
        Vuc, Vus = self.Vu(xu)
        Vu = Vuc + sum(Vus)
        s += f"{'Vu = ':>5}{Vu/1e3:.2f} kN\n"
        return s

    def eff_d(self, xu: float) -> float:
        _, ct = self.long_steel.centroid(xu)
        return ct

    def pt(self, xu: float) -> float:
        ast = 0.0
        for L in sorted(self.long_steel.layers):
            if L._xc > xu:
                ast += L.area
        d = self.eff_d(xu)
        pt = ast / (self.b * d) * 100
        return pt

    def Vu(self, xu: float) -> Tuple[float, List[float]]:
        # print("\nstart::RectBeamSection.Vu(xu)", xu)
        pt = self.pt(xu)
        # print("stop::RectBeamSection.Vu(xu)\n")
        tauc = self.conc.tauc(pt)
        d = self.eff_d(xu)
        vuc = tauc * self.b * d
        vus = self.shear_steel.Vus(d)
        return vuc, vus

    def analyse(self, ecmax: float = ecu) -> Tuple[float, float]:
        xu = self.xu(ecmax)
        Mu = self.Mu(xu, ecmax)
        return xu, Mu

    def design_singly(self, bar_dia: float, Mu: float) -> Tuple[float, float]:
        beam = Beam()
        fck = self.conc.fck
        bottom_layer = self.long_steel.layers[-1]
        fy = bottom_layer.rebar.fy
        fd = bottom_layer.rebar.fd
        d = self.D - self.clear_cover - bar_dia / 2
        dc = self.clear_cover + bar_dia / 2
        Mulim = beam.Mulim_const(fy) * fck * self.b * d**2
        if Mu < Mulim:
            ast = beam.reqd_Ast(fck, fy, self.b, d, Mu)
            asc = 0.0
        else:
            ast1 = beam.reqd_Ast(fck, fy, self.b, d, Mulim)
            Mu2 = Mu - Mulim
            ast2 = Mu2 / (fd * (d - dc))
            ast = ast1 + ast2
            xu = beam.xumax_d(fy) * d
            esc = self.csb.ecu / xu * (xu - dc)
            fsc = bottom_layer.rebar.fs(esc)
            fcc = self.csb._fc_(esc) * self.conc.fd
            asc = ast2 * fd / (fsc - fcc)
            print("---", xu, d, esc, fsc, fcc, ast1, ast2, ast, asc)
        return ast, asc


"""Class to repersent flanged section"""


class FlangedBeamSection(RectBeamSection):
    def __init__(
        self,
        bw: float,
        D: float,
        bf: float,
        Df: float,
        csb: LSMStressBlock,
        conc: Concrete,
        long_steel: RebarGroup,
        shear_steel: ShearRebarGroup,
        clear_cover: float,
    ):
        super().__init__(bw, D, csb, conc, long_steel, shear_steel, clear_cover)
        self.bf = bf
        self.Df = Df

    @property
    def bw(self) -> float:
        return self.b

    @bw.setter
    def bw(self, _bw) -> None:
        self.b = _bw

    def Cw(self, xu: float, ecmax: float = ecu) -> Tuple[float, float]:
        k = xu / self.D
        area = self.csb.C(0, k, k, ecmax) * self.conc.fd
        moment = self.csb.M(0, k, k, ecmax) * self.conc.fd

        C = area * self.bw * self.D
        M = moment * self.bw * self.D**2
        return C, M

    def Cf(self, xu: float, ecmax: float = ecu) -> Tuple[float, float]:
        df = xu if xu <= self.Df else self.Df
        k = xu / self.D
        z1 = (xu - df) / self.D
        z2 = k
        area = self.csb.C(z1, z2, k, ecmax) * self.conc.fd
        moment = self.csb.M(z1, z2, k, ecmax) * self.conc.fd
        C = area * self.D * (self.bf - self.bw)
        M = moment * self.D**2 * (self.bf - self.bw)
        return C, M

    def C_M(self, xu: float, ecmax: float = ecu) -> Tuple[float, float]:
        # Compression force and moment due to concrete of web
        C1, M1 = self.Cw(xu, ecmax)
        C2, M2 = self.Cf(xu, ecmax)
        # Compression force and moment due to compression reinforcement bars
        if self.has_compr_steel(xu):
            C3, M3 = self.long_steel.force_compression(xu, self.csb, self.conc, ecmax)
        else:
            C3 = M3 = 0.0

        # Sum it all up
        C = C1 + C2 + C3
        M = M1 + M2 + M3
        return C, M

    def Mu(self, xu: float, ecmax: float = ecu) -> float:
        # Cc and Ct must be equal for beams, if not, NA chosen does not conform to equilibrium
        Cc, Mc = self.C_M(xu, ecmax)
        Ct, Mt = self.T(xu, ecmax)
        Mu = Mc + Mt
        return Mu

    def __repr__(self) -> str:
        ecmax = self.csb.ecu
        xu = self.xu(ecmax)
        return self.report(xu, ecmax)

    def C_T(self, xu: float, ecmax: float = ecu) -> float:
        C, _ = self.C_M(xu, ecmax)
        T, _ = self.T(xu, ecmax)
        return C - T

    def xu(self, ecmax: float = ecu) -> float | Any:
        x1, x2 = rootsearch(self.C_T, 10, self.D, 10, ecmax)
        x = brentq(self.C_T, x1, x2, args=(ecmax,))
        return x

    def report(self, xu: float, ecmax: float = ecu) -> str:  # pragma: no cover
        self.calc_xc()
        self.calc_stress_type(xu)
        k = xu / self.D
        ecy = self.csb.ecy
        hdr0 = f"FLANGED BEAM SECTION - Web: {self.b} x {self.D}, Flange: {self.bf} x {self.Df}"
        s = f"{header(hdr0, '~')}\n"
        s += f"{header('FLEXURE', '=')}\nEquilibrium NA = {xu:.2f} (ec_max = {ecmax:.6f})\n\n"
        Fcw, Mcw = self.Cw(xu, ecmax)
        Fcf, Mcf = self.Cf(xu)
        Mc = Mcw + Mcf
        hdr1 = f"{'fck':>6} {'Breadth':>10} {'Depth':>10} {'ec_min':>12}  {'ec_max':>12} {'Type':>6} "
        hdr1 += f"{'C (kN)':>8} {'M (kNm)':>8}"
        # Web
        s += hdr1 + "\n" + underline(hdr1) + "\n"
        s += f"{self.conc.fck:6.0f} {self.bw:10.2f} {self.D:10.2f} {0:12.8f}  {ecmax:12.8f} {'C':>6} "
        s += f"{Fcw/1e3:8.2f} {Mcw/1e6:8.2f}\n"
        # Flange
        s += f"{self.conc.fck:6.0f} {self.bf:10.2f} {self.Df:10.2f} {' ':>12}  {ecmax:12.8f} {'C':>6} "
        s += f"{Fcf/1e3:8.2f} {Mcf/1e6:8.2f}\n"
        s += f"{underline(hdr1)}\n"
        s += f"{' ':>62} {(Fcw+Fcf)/1e3:8.2f} {(Mcw+Mcf)/1e6:8.2f}\n"
        hdr2 = f"{'fy':>6} {'Bars':>12} {'xc':>8} {'Strain':>12} {'Type':>4} {'f_s':>8} {'f_c':>6}"
        hdr2 += f" {'C (kN)':>8} {'M (kNm)':>8}"
        s += f"\n{hdr2}\n{underline(hdr2)}\n"
        Ft = 0.0
        Mt = 0.0
        for L in sorted(self.long_steel.layers):
            z = k - (L._xc / self.D)
            esc = self.csb.ec(z, k) * ecy
            stress_type = L.stress_type(xu)
            fsc = L.rebar.fs(esc)
            s += f"{L.rebar.fy:6.0f} {L.bar_list():>12} {L._xc:8.2f} {esc:12.8f} "
            s += f"{StressLabel[stress_type][0]:>4} {fsc:8.2f} "
            if stress_type == StressType.STRESS_COMPRESSION:
                fcc = self.csb.fc(z, k, ecmax) * self.conc.fd
                c = L.area * (fsc - fcc)
                s += f"{fcc:6.2f} "
            else:
                c = L.area * fsc
                s += f"{' ':>6} "
            m = c * (k * self.D - L._xc)
            s += f"{c/1e3:8.2f} {m/1e6:8.2f}\n"
            Ft += c
            Mt += m
        s += f"{underline(hdr2)}\n"
        if len(self.long_steel.layers) > 1:
            C_M = f"{Ft/1e3:8.2f} {Mt/1e6:8.2f}"
            s += f"{' ':>62} {C_M}\n{' ':>62} {underline(C_M, '=')}\n"
        F = 0.0 if isclose(Fcw + Fcf + Ft, 0, abs_tol=1e-10) else Fcw + Fcf + Ft
        s += f"{' ':>62} {F/1e3:8.2f} {(Mc + Mt)/1e6:8.2f}\n"

        s += f"{header('SHEAR', '=')}\n"
        tauc = self.conc.tauc(self.pt(xu))
        area = self.b * self.eff_d(xu)
        vuc = area * tauc
        hdr3 = f"{'Type':>14} {' ':>14} {'tau_c':>6} {'Area (mm^2)':>16} {' ':>8} {' ':>8} {'V_uc (kN)':>8}"
        s += f"{header(hdr3)}\n"
        s += f"{'Concrete':>14} {' ':>14} {tauc:6.2f} {area:16.2f} {' ':>8} {' ':>8} {vuc/1e3:8.2f}\n"
        s += f"{underline(hdr3)}\n"
        hdr4 = f"{'Type':>14} {'Variant':>14} {'f_y':>6} {'Bars':>16} {'s_v':>8} {'A_sv':>8} {'V_us (kN)':>8}"
        s += f"{header(hdr4)}\n"
        vus = 0.0
        for sh_rein in self.shear_steel.shear_reinforcement:
            data = sh_rein.report(self.eff_d(xu))
            s += f"{data['label']:>14} {data['type']:>14} {data['fy']:6} "
            if data["sh_type"] in [
                ShearRebarType.SHEAR_REBAR_VERTICAL_STIRRUP,
                ShearRebarType.SHEAR_REBAR_INCLINED_STIRRUP,
            ]:
                bar_info = f"{data['legs']}-{data['bar_dia']}#"
            else:
                bar_info = f"{data['bars']}"
            s += f"{bar_info:>16} {data['sv']:8.1f} {data['Asv']:8.2f} {data['Vus']/1e3:8.2f}\n"
            vus += data["Vus"]
        vu = f"{(vuc + vus)/1e3:8.2f}"
        s += f"{' ':>71} {underline(vu, '=')}\n{' ':>71} {vu}\n"
        s += (
            f"{header('CAPACITY', '=')}\n{'Mu = ':>5}{self.Mu(xu, ecmax)/1e6:.2f} kNm\n"
        )
        Vuc, Vus = self.Vu(xu)
        Vu = Vuc + sum(Vus)
        s += f"{'Vu = ':>5}{Vu/1e3:.2f} kN\n"
        return s


class RectColumnSection:
    def __init__(
        self,
        b: float,
        D: float,
        csb: LSMStressBlock,
        conc: Concrete,
        long_steel: RebarGroup,
        lat_ties: LateralTie,
        clear_cover: float,
    ):
        self.design_force_type = DesignForceType.COLUMN
        self.csb = csb
        self.conc = conc
        self.long_steel = long_steel
        self.clear_cover = clear_cover
        self.b = b
        self.D = D
        self.long_steel = long_steel
        self.lat_ties = lat_ties

    @property
    def Asc(self) -> float:
        return self.long_steel.area

    def k(self, xu: float) -> float:
        return xu / self.D

    def C_M(self, xu: float) -> Tuple[float, float]:
        self.long_steel.calc_xc(self.D)
        k = self.k(xu)
        if xu <= self.D:
            z1 = 0.0
            z2 = k
        else:
            z1 = k - 1
            z2 = k
        a = self.csb.C(z1, z2, k)
        Cc = a * self.conc.fd * self.b * self.D
        m = self.csb.M(z1, z2, k)
        Mc = m * self.conc.fd * self.b * self.D**2
        Cs = 0.0
        Ms = 0.0
        for L in self.long_steel.layers:
            asc = L.area
            x = xu - L.xc
            z = x / self.D
            esc = self.csb.ec(z, k) * self.csb.ecy
            fsc = L.rebar.fs(esc)
            fcc = self.csb.fc(z, k) * self.conc.fd
            if fsc >= 0:
                _Cs = asc * (fsc - fcc)
            else:
                _Cs = asc * fsc
            Cs += _Cs
            Ms += _Cs * abs(x)
        return Cc + Cs, Mc + Ms

    def __repr__(self) -> str:
        s = f"RECTANGULAR COLUMN {self.b} x {self.D}\n"
        s += f"Concrete: {self.conc} Clear Cover: {self.clear_cover}\n"
        self.long_steel.calc_xc(self.D)
        s += f"{'fy':>6} {'Bars':>8} {'xc':>8}\n"
        for L in self.long_steel.layers:
            s += f"{L.rebar.fy:6.0f} {L.bar_list():>8} {L._xc:8.2f}\n"
        return s

    def report(self, xu: float) -> str:
        k = xu / self.D
        ecy = self.csb.ecy
        hdr0 = f"RECTANGULAR COLUMN {self.b} x {self.D} xu = {xu:.2f} (k = {k:.2f})"
        s = f"{header(hdr0, '~')}\n"
        s += f"Concrete: {self.conc} Clear Cover: {self.clear_cover}\n"
        # Concrete
        fd = self.conc.fd
        if k <= 1:
            z1 = 0.0
        else:
            z1 = k - 1
        z2 = k
        ecmin = self.csb.ec(z1, k) * ecy
        ecmax = self.csb.ec(z2, k) * ecy
        fsc1 = self.csb.fc(z1, k) * fd
        fsc2 = self.csb.fc(z2, k) * fd
        Cc = self.csb.C(z1, z2, k) * fd * self.b * self.D
        Mc = self.csb.M(z1, z2, k) * fd * self.b * self.D**2
        hdr1 = f"{'fck':>6} {' ':>8} {'ecmin':>12} {'ecmax':>12} {'Type':>4} {'fsc1':>8} {'fsc2':>6} "
        hdr1 += f"{'Cc':>8} {'Mc':>8}"
        s += f"\n{header(hdr1)}\n"
        # s += header(hdr1) + "\n"
        s += f"{self.conc.fck:6.2f} {' ':>8} {ecmin:12.8f} {ecmax:12.8f} {'C':>4} {fsc1:8.2f} {fsc2:6.2f} "
        s += f"{Cc/1e3:8.2f} {Mc/1e6:8.2f}\n{'-'*len(hdr1)}\n"
        # Longitudinal steel
        self.long_steel.calc_xc(self.D)
        self.long_steel.calc_stress_type(xu)
        hdr2 = f"{'fy':>6} {'Bars':>12} {'xc':>8} {'Strain':>12} {'Type':>4} {'fsc':>8} {'fcc':>6} "
        hdr2 += f"{'C (kN)':>8} {'M (kNm)':>8}"
        s += f"\n{header(hdr2)}\n"
        ecy = self.csb.ecy
        cc = 0.0
        mm = 0.0
        for L in sorted(self.long_steel.layers):
            z = k - (L._xc / self.D)
            esc = self.csb.ec(z, k) * ecy
            str_type = L.stress_type(xu)
            fsc = L.rebar.fs(esc)
            if str_type == 1:
                fcc = self.csb.fc(z, k) * self.conc.fd
                c = L.area * (fsc - fcc)
            else:
                c = L.area * fsc
                fcc = 0.0
            m = c * (k * self.D - L._xc)
            s += f"{L.rebar.fy:6.0f} {L.bar_list():>12} {L._xc:8.2f} {esc:12.8f} {StressLabel[str_type][0]:>4} "
            if str_type == 1:
                s += f"{fsc:8.2f} {fcc:6.2f}"
            else:
                s += f"{fsc:8.2f} {'--':>6}"
            s += f" {c/1e3:8.2f} {m/1e6:8.2f}\n"
            cc += c
            mm += m
        s += "-" * len(hdr2) + "\n"
        s += f"{' '*62} {cc/1e3:8.2f} {mm/1e6:8.2f}\n"
        C, M = self.C_M(xu)
        hdr3 = f"{C/1e3:8.2f} {M/1e6:8.2f}"
        s += f"{' ':>62} {underline(hdr3, '=')}\n{' ':>62} {hdr3}\n"
        s += f"{header('CAPACITY', '=')}\n"
        s += f"Pu = {C/1e3:10.2f} kN\nMu = {M/1e6:10.2f} kNm\n e = {M/C:10.2f} mm\n"
        return s



from typing import Any
from sympy import symbols, nsimplify, integrate
from sympy.core.mul import Mul

from rcdesign.is456 import ecy, ecu

# from rcdesign.stressblock import StressBlock


class LSMStressBlock:
    def __init__(self, label: str = "IS 456 LSM", ecy: float = ecy, ecu: float = ecu):
        self.label = label
        self.ecy = ecy
        self.ecu = ecu

    def __repr__(self):
        s = self.label
        return s

    def isvalid_ecmax(self, ecmax: float) -> float:
        if (ecmax < 0) or (ecmax > self.ecu):
            raise ValueError
        else:
            return ecmax

    def isvalid_k(self, k: float) -> float:
        if k < 0:
            raise ValueError
        else:
            return k

    def isvalid_z(self, z: float, k: float):
        k = self.isvalid_k(k)
        if (z < k - 1) or (z > k):
            raise ValueError
        else:
            return z

    def _ec(self, _k: float, ecmax: float = ecu) -> Mul | Any:
        ecmax = self.isvalid_ecmax(ecmax)
        z, k = symbols("z k")
        if _k < 0:  # Invalid values for k
            raise ValueError
        elif _k == 0:  # Unstressed condition
            ec = nsimplify(0)
        elif _k <= 1:  # Flexure or axial compression with NA within the section
            ec = (z / nsimplify(self.ecy / ecmax * k)).evalf(subs={"k": _k})
        else:  # Axial compression with NA outside the section
            ec = (z / nsimplify((k - 3 / 7))).evalf(subs={"k": _k})
        return ec

    def ec(self, z: float, k: float, ecmax: float = ecu) -> float:
        _ec = self._ec(k, ecmax)
        return float(_ec.evalf(subs={"z": z}))

    def _fc_(self, ec: float) -> float:
        if (ec < 0) or (ec > self.ecu):
            return 0.0
        ec_ecy = ec / self.ecy
        if ec_ecy < 1:
            return 2 * ec_ecy - ec_ecy**2
        else:
            return 1.0

    def _fc(self, z: float, k: float, ecmax: float = ecu) -> Mul | Any:
        ecmax = self.isvalid_ecmax(ecmax)
        if k < 0:  # Invalid values for k
            raise ValueError
        elif k == 0:  # Unstressed condition
            fc = nsimplify(0)
        else:  # Flexure or axial compression with NA within the section
            ec = self.ec(z, k, ecmax)
            if ec < 0:
                fc = nsimplify(0)
            elif ec < 1:
                ec_ecy = self._ec(k, ecmax)
                fc = (2 * ec_ecy - ec_ecy**2).evalf(subs={"k": k})
            else:
                fc = nsimplify(1)
        return fc

    def fc(self, z: float, k: float, ecmax: float = ecu) -> float:
        fc = self._fc(z, k, ecmax)
        return float(fc.evalf(subs={"z": z}))

    def C(self, z1: float, z2: float, k: float, ecmax: float = ecu) -> float | Any:
        ecmax = self.isvalid_ecmax(ecmax)
        z = symbols("z")
        k = self.isvalid_k(k)
        if k == 0:
            return 0.0

        z1 = self.isvalid_z(z1, k)
        z2 = self.isvalid_z(z2, k)
        if z1 > z2:
            z1, z2 = z2, z1
        ec1 = self.ec(z1, k)
        ec2 = self.ec(z2, k)
        if ec2 <= 1:  # Parabolic only
            fcp = self._fc(z1, k, ecmax)
            Cp = integrate(fcp, (z, z1, z2))
            Cr = 0.0
        elif ec1 >= 1:  # Rectangular only
            Cp = 0.0
            fcr = self._fc(z1, k, ecmax)
            Cr = integrate(fcr, (z, z1, z2))
        else:  # Both Parabolic and Rectangular
            if k <= 1:  # NA within the section
                zz = self.ecy / ecmax * k
            else:  # NA outside the section
                zz = k - (1 - self.ecy / self.ecu)
            fcp = self._fc(z1, k, ecmax)
            fcr = self._fc(z2, k, ecmax)
            Cp = integrate(fcp, (z, z1, zz))
            Cr = integrate(fcr, (z, zz, z2))
        return Cp + Cr

    def M(self, z1: float, z2: float, k: float, ecmax: float = ecu) -> float | Any:
        ecmax = self.isvalid_ecmax(ecmax)
        z = symbols("z")
        k = self.isvalid_k(k)
        if k == 0:
            return 0.0

        z1 = self.isvalid_z(z1, k)
        z2 = self.isvalid_z(z2, k)
        if z1 > z2:
            z1, z2 = z2, z1
        ec1 = self.ec(z1, k)
        ec2 = self.ec(z2, k)
        if ec2 <= 1:  # Parabolic only
            fcp = self._fc(z1, k, ecmax)
            Mp = integrate(fcp * z, (z, z1, z2))
            Mr = 0.0
        elif ec1 >= 1:  # Rectangular only
            Mp = 0.0
            fcr = self._fc(z1, k, ecmax)
            Mr = integrate(fcr * z, (z, z1, z2))
        else:  # Both Parabolic and Rectangular
            if k <= 1:  # NA within the section
                zz = self.ecy / ecmax * k
            else:  # NA outside the section
                zz = k - (1 - self.ecy / self.ecu)
            fcp = self._fc(z1, k, ecmax)
            fcr = self._fc(z2, k, ecmax)
            Mp = integrate(fcp * z, (z, z1, zz))
            Mr = integrate(fcr * z, (z, zz, z2))
        return Mp + Mr



# import numpy as np
from numpy import sign

# from scipy.optimize import brentq
from typing import Callable


def underline(s: str, ch: str = "-") -> str:
    return ch * len(s)


def header(s: str, ch: str = "-") -> str:
    return s + "\n" + underline(s, ch)


def func(x: float, *args: float) -> float:
    # return a*x**2 + c*x - 4
    # return x**3 - 10*x**2 + 5
    return args[0] ** 3 + args[1] * x**2 + args[2] * x + args[3]


def rootsearch(func: Callable, xstart: float, xstop: float, numint: int, *args):
    dx = (xstop - xstart) / (numint)
    x1 = xstart
    y1 = func(x1, *args)
    while x1 < xstop:
        x2 = x1 + dx
        y2 = func(x2, *args)
        if sign(y1) != sign(y2):
            return (x1, x2)
        else:
            x1, y1 = x2, y2
    return (None, None)


def ceiling(x: float, multipleof: float = 1.0):
    i = x // multipleof
    if (x - i * multipleof) > 0:
        i += 1
    return i * multipleof


def floor(x: float, multipleof: float = 1.0):
    i = x // multipleof
    return i * multipleof
